@@ -3,7 +3,7 @@
  * Mark Nagel <nagel@ics.uci.edu>
  * 20 July 1989
  *
- * $Revision: 7.15 $
+ * $Revision: 7.16 $
  *
  * bool
  * format(fmt, precision, num, buf, buflen)
@@ -258,11 +258,11 @@ format(char *fmt, int lprecision, double val, char *buf, int buflen)
     }
     (void) sprintf(prtfmt, "%%.%dlf", width);
     (void) sprintf(mantissa, prtfmt, val);
-    for (cp = integer = mantissa; *cp != '.' && *cp != EOS; cp++) {
+    for (cp = integer = mantissa; *cp != dpoint && *cp != EOS; cp++) {
 	if (*integer == '0')
 	    integer++;
     }
-    if (*cp == '.') {
+    if (*cp == dpoint) {
 	fraction = cp + 1;
 	*cp = EOS;
 	cp = fraction + strlen(fraction) - 1;
@@ -352,8 +352,9 @@ fmt_int(char *val,	/* integer part of the value to be formatted */
 	} else if (f >= 0 && (fmt[f] == '#' || fmt[f] == '0')) {
 	    if (v >= 0 || fmt[f] == '0') {
 		*bufptr++ = v < 0 ? '0' : val[v];
-		if (comma && (thousands = (thousands + 1) % 3) == 0 && v > 0)
-		    *bufptr++ = ',';
+		if (comma && (thousands = (thousands + 1) % 3) == 0 &&
+			v > 0 && thsep != '\0')
+		    *bufptr++ = thsep;
 		v--;
 	    }
 	} else if (f >= 0) {
@@ -384,7 +385,7 @@ fmt_frac(char *val,	/* fractional part of the value to be formatted */
     register char *bufptr = buf;
     register char *fmtptr = fmt, *valptr = val;
 
-    *bufptr++ = '.';
+    *bufptr++ = dpoint;
     while (*fmtptr != EOS) {
 	if (*fmtptr == '&') {
 	    int i;
