@@ -4,7 +4,7 @@
  *
  *              Robert Bond, 4/87
  *
- *		$Revision: 7.15 $
+ *		$Revision: 7.16 $
  */
 
 #include <sys/types.h>
@@ -73,9 +73,16 @@ add_range(char *name, struct ent_ptr left, struct ent_ptr right, int is_range)
     p = name;
     if (isdigit(*p) || (isalpha(*p++) && (isdigit(*p) ||
 		(isalpha(*p++) && isdigit(*p))))) {
-	p++;
-	while (isdigit(*p))
-	    p++;
+	if (*name == '0' && (name[1] == 'x' || name[1] == 'X')) {
+	    ++p;
+	    while (isxdigit(*++p)) /* */;
+	    if (*p == 'p' || *p == 'P')
+		while (isxdigit(*++p)) /* */;
+	} else {
+	    while (isdigit(*++p)) /* */;
+	    if (isdigit(*name) && (*p == 'e' || *p == 'E'))
+		while (isdigit(*++p)) /* */;
+	}
 	if (!(*p)) {
 	    error("Invalid range name \"%s\" - ambiguous", name);
 	    scxfree(name);
