@@ -7,7 +7,7 @@
  *
  *              More mods Robert Bond, 12/86
  *		More mods by Alan Silverstein, 3-4/88, see list of changes.
- *		$Revision: 7.15 $
+ *		$Revision: 7.16 $
  *
  */
 
@@ -16,7 +16,6 @@
 #include <signal.h>
 #include <curses.h>
 #include <ctype.h>
-#include <getopt.h>
 
 #ifdef BSD42
 #include <strings.h>
@@ -41,12 +40,6 @@
 #define	SAVENAME "SC.SAVE" /* file name to use for emergency saves */
 #endif /* SAVENAME */
 
-#ifndef DFLT_PAGER
-#define	DFLT_PAGER "more"	/* more is probably more widespread than less */
-#endif /* DFLT_PAGER */
-
-#define MAXCMD 160	/* for ! command below */
-
 /* Globals defined in sc.h */
 
 struct ent ***tbl;
@@ -66,8 +59,9 @@ char *row_hidden;
 char line[FBUFLEN];
 int changed;
 struct ent *delbuf[DELBUFSIZE];
+char *delbuffmt[DELBUFSIZE];
 int dbidx;
-int qbuf;	/* buffer no specified by " command */
+int qbuf;	/* buffer no. specified by " command */
 int modflg;
 int cellassign;
 int numeric;
@@ -90,12 +84,22 @@ int brokenpipe = FALSE;	/* Set to true if SIGPIPE is received */
 int	wasforw	= FALSE;
 #endif
 
-void	update();
-void	repaint();
+void		update();
+void		repaint();
+extern void	doshell();
+extern void	gohome();
+extern void	leftlimit();
+extern void	rightlimit();
+extern void	gototop();
+extern void	gotobottom();
 
-char curfile[PATHLEN];
+char    curfile[PATHLEN];
 char    revmsg[80];
 
+/* numeric separators, country-dependent if locale support enabled: */
+char	dpoint = '.';	/* decimal point */
+char	thsep = ',';	/* thousands separator */
+
 int  linelim = -1;
 
 int  showtop   = 1;	/* Causes current cell value display in top line  */
@@ -113,6 +117,10 @@ int  autolabel = 1;     /* If room, causes label to be created after a define */
 int  autoinsert = 0;    /* Causes rows to be inserted if craction is non-zero
 			   and the last cell in a row/column of the scrolling
 			   portion of a framed range has been filled	  */
+int  autowrap = 0;      /* Causes cursor to move to next row/column if craction
+			   is non-zero and the last cell in a row/column of
+			   the scrolling portion of a framed range has been
+			   filled */
 int  calc_order = BYROWS;
 int  optimize  = 0;	/* Causes numeric expressions to be optimized */
 int  tbl_style = 0;	/* headers for T command output */
@@ -154,16 +162,16 @@ lookat(int row, int col)
 
     checkbounds(&row, &col);
     pp = ATBL(tbl, row, col);
-    if (*pp == (struct ent *)0) {
+    if (*pp == NULL) {
         if (freeents != NULL) {
 	    *pp = freeents;
-	    (*pp)->flags &= ~(is_cleared);
+	    (*pp)->flags &= ~is_cleared;
 	    (*pp)->flags |= may_sync;
 	    freeents = freeents->next;
 	} else
 	    *pp = (struct ent *) scxmalloc((unsigned)sizeof(struct ent));
-	if (row>maxrow) maxrow = row;
-	if (col>maxcol) maxcol = col;
+	if (row > maxrow) maxrow = row;
+	if (col > maxcol) maxcol = col;
 	(*pp)->label = (char *)0;
 	(*pp)->row = row;
 	(*pp)->col = col;
@@ -186,12 +194,13 @@ lookat(int row, int col)
  * We also use it as a last-deleted buffer for the 'p' command.
  */
 void
-free_ent(register struct ent *p)
+free_ent(register struct ent *p, int unlock)
 {
     p->next = delbuf[dbidx];
     delbuf[dbidx] = p;
     p->flags |= is_deleted;
-    p->flags &= ~is_locked;
+    if (unlock)
+	p->flags &= ~is_locked;
 }
 
 /* free deleted cells */
@@ -203,7 +212,10 @@ flush_saved()
 
     if (dbidx < 0)
 	return;
-    (p = delbuf[dbidx]);
+    if (p = delbuf[dbidx]) {
+	scxfree(delbuffmt[dbidx]);
+	delbuffmt[dbidx] = NULL;
+    }
     while (p) {
 	(void) clearent(p);
 	q = p->next;
@@ -262,7 +274,7 @@ main (int argc, char  **argv)
     else
 	progname = argv[0];
 
-    while ((c = getopt(argc, argv, "axmoncrCReP:vq")) != EOF) {
+    while ((c = getopt(argc, argv, "axmoncrCReP:W:vq")) != EOF) {
     	switch (c) {
 	    case 'a':
 		    skipautorun = 1;
@@ -303,6 +315,7 @@ main (int argc, char  **argv)
 		    eopt = 1;
 		    break;
 	    case 'P':
+	    case 'W':
 		    popt = 1;
 	    case 'v':
 		    break;
@@ -310,8 +323,8 @@ main (int argc, char  **argv)
 		    qopt = 1;
 		    break;
 	    default:
-		    (void) fprintf(stderr,"Usage: %s [-acemnoqrxCR] [file...]\n",
-			progname);
+		    (void) fprintf(stderr,
+			    "Usage: %s [-acemnoqrxCR] [file...]\n", progname);
 		    exit (1);
 	}
     }
@@ -319,6 +332,7 @@ main (int argc, char  **argv)
     if (!isatty(STDOUT_FILENO) || popt || qopt) usecurses = FALSE;
     startdisp();
     signals();
+    read_hist();
 
     /* setup the spreadsheet arrays, initscr() will get the screen size */
     if (!growtbl(GROWNEW, 0, 0)) {
@@ -346,21 +360,29 @@ main (int argc, char  **argv)
     if (optind < argc && !strcmp(argv[optind], "--"))
 	optind++;
     if (optind < argc && argv[optind][0] != '|' &&
-	    !strcmp(argv[optind], "-"))
+	    strcmp(argv[optind], "-"))
 #endif /* MSDOS */
 	(void) strcpy(curfile, argv[optind]);
-    for (dbidx = DELBUFSIZE - 1; dbidx >= 0; )
-	delbuf[dbidx--] = NULL;
+    for (dbidx = DELBUFSIZE - 1; dbidx >= 0; ) {
+	delbuf[dbidx] = NULL;
+	delbuffmt[dbidx--] = NULL;
+    }
+    if (usecurses && has_colors())
+	initcolor(0);
+
     if (optind < argc) {
-	readfile(argv[optind], 1);
+	if (!readfile(argv[optind], 1) && (optind == argc - 1))
+	    error("New file: \"%s\"", curfile);
 	EvalAll();
 	optind++;
     } else
 	erasedb();
+
     while (optind < argc) {
-	readfile(argv[optind], 0);
+	(void) readfile(argv[optind], 0);
 	optind++;
     }
+
     savedrow[0] = currow;
     savedcol[0] = curcol;
     savedstrow[0] = strow;
@@ -368,16 +390,13 @@ main (int argc, char  **argv)
     EvalAll();
 
     if (!(popt || isatty(STDIN_FILENO)))
-	readfile("-", 0);
+	(void) readfile("-", 0);
 
     if (qopt) {
 	stopdisp();
 	exit (0);
     }
 
-    if (usecurses && has_colors())
-	initcolor(0);
-
     clearok(stdscr, TRUE);
     EvalAll();
 
@@ -401,9 +420,12 @@ main (int argc, char  **argv)
 	char *redraw = NULL;
 	int o;
 
-	optind = 0;
+#ifdef BSD43
+	optreset = 1;
+#endif
+	optind = 1;
 	stopdisp();
-	while ((o = getopt(argc, argv, "xmoncrCReP:vq")) != EOF) {
+	while ((o = getopt(argc, argv, "axmoncrCReP:W:vq")) != EOF) {
 	    switch (o) {
 		case 'v':
 		    Vopt = 1;
@@ -457,7 +479,7 @@ main (int argc, char  **argv)
 				    break;
 				case ctl('l'):
 				    FullUpdate++;
-				    clearok(stdscr,1);
+				    clearok(stdscr, 1);
 				    break;
 				default:
 				    write_line(c);
@@ -488,6 +510,12 @@ main (int argc, char  **argv)
 		    }
 		    Vopt = 0;
 		    break;
+		case 'W':
+		    strcpy(line, "write ");
+		    strcat(line, optarg);
+		    linelim = 0;
+		    yyparse();
+		    break;
 		default:
 		    break;
 	    }
@@ -726,7 +754,6 @@ main (int argc, char  **argv)
 
 		case ctl('m'):
 		case ctl('j'):
-		    numeric_field = 0;
 		    write_line(ctl('m'));
 		    break;
 
@@ -902,6 +929,11 @@ main (int argc, char  **argv)
 			    error("Autoinsert %sabled.",
 				   autoinsert? "en" : "dis");
 			    break;
+			case 'w': case 'W':
+			    autowrap = (!autowrap);
+			    error("Autowrap %sabled.",
+				   autowrap? "en" : "dis");
+			    break;
 			case 'z': case 'Z':
 			    rowlimit = currow;
 			    collimit = curcol;
@@ -963,6 +995,10 @@ main (int argc, char  **argv)
 			FullUpdate++;
 		    }
 		    break;
+		case '\035':	/* ^] */
+		    if (linelim >= 0)
+			write_line(c);
+		    break;
 
 	    } /* End of the control char switch stmt */
 	else if (isascii(c) && isdigit(c) && ((!numeric && linelim < 0) ||
@@ -974,13 +1010,8 @@ main (int argc, char  **argv)
 		if (c == '0') {    /* just a '0' goes to left col */
 		    if (linelim >= 0)
 			write_line(c);
-		    else {
-			remember(0);
-			curcol = 0;
-			rowsinrange = 1;
-			colsinrange = fwidth[curcol];
-			remember(1);
-		    }
+		    else
+			leftlimit();
 		} else {
 		    nedistate = 0;
 		    narg = c - '0';
@@ -994,6 +1025,7 @@ main (int argc, char  **argv)
 	    deraw(1);
 	    system("man sc");
 	    goraw();
+	    clear();
 	} else if (linelim >= 0) {
 	    /* Editing line */
 	    switch (c) {
@@ -1010,8 +1042,15 @@ main (int argc, char  **argv)
 	} else if (!numeric && ( c == '+' || c == '-' )) {
 	    /* increment/decrement ops */
 	    register struct ent *p = *ATBL(tbl, currow, curcol);
-	    if (!p || !(p->flags & is_valid))
+	    if (!p || !(p->flags & is_valid)) {
+		if (c == '+') {
+		    editv(currow, curcol);
+		    linelim = strlen(line);
+		    insert_mode();
+		    write_line(ctl('v'));
+		}
 		continue;
+	    }
 	    if (p->expr && !(p->flags & is_strexpr)) {
 		error("Can't increment/decrement a formula\n");
 		continue;
@@ -1027,8 +1066,9 @@ main (int argc, char  **argv)
 	    if (fkey[c - KEY_F0 - 1]) {
 		char *tpp;
 
-		write_line(ctl('m'));	/* switch to insert mode */
+		insert_mode();
 		strcpy(line, fkey[c - KEY_F0 - 1]);
+		linelim = 0;
 		for (tpp = line; *tpp != '\0'; tpp++)
 		    if (*tpp == '\\' && *(tpp + 1) == '"')
 			memmove(tpp, tpp + 1, strlen(tpp));
@@ -1088,12 +1128,12 @@ main (int argc, char  **argv)
 
 			numeric_field = 1;
 			editv(currow, curcol);
+			linelim = strlen(line);
 			insert_mode();
-			if (p->flags & is_valid) {
-			    write_line(ESC);
-			    write_line('a');
-			}
-			write_line(c);
+			if (c == '-' || p->flags & is_valid)
+			    write_line(c);
+			else
+			    write_line(ctl('v'));
 		    }
 		    break;
 
@@ -1112,63 +1152,14 @@ main (int argc, char  **argv)
 		    break;
 
 		case '!':
-		    {
-		    /*
-		     *  "! command"  executes command
-		     *  "!"	forks a shell
-		     *  "!!" repeats last command
-		     */
-#if VMS || MSDOS
-		    error("Not implemented on VMS or MS-DOS");
-#else /* VMS */
-		    char *shl;
-		    int pid, temp;
-		    char cmd[MAXCMD];
-		    static char lastcmd[MAXCMD];
-
-		    if (!(shl = getenv("SHELL")))
-			shl = "/bin/sh";
-
-		    deraw(1);
-		    (void) fputs("! ", stdout);
-		    (void) fflush(stdout);
-		    (void) fgets(cmd, MAXCMD, stdin);
-		    cmd[strlen(cmd) - 1] = '\0';	/* clobber \n */
-		    if (strcmp(cmd,"!") == 0)		/* repeat? */
-			(void) strcpy(cmd, lastcmd);
-		    else
-			(void) strcpy(lastcmd, cmd);
-
-		    if (modflg) {
-			(void) puts ("[No write since last change]");
-			(void) fflush(stdout);
-		    }
-
-		    if (!(pid = fork())) {
-			(void) signal(SIGINT, SIG_DFL);  /* reset */
-			if(strlen(cmd))
-			    (void) execl(shl,shl,"-c",cmd,(char *)0);
-			else
-			    (void) execl(shl, shl, (char *)0);
-			exit (-127);
-		    }
-
-		    while (pid != wait(&temp));
-
-		    (void) printf("Press any key to continue ");
-		    fflush(stdout);
-		    cbreak();
-		    (void)getch();
-		    goraw();
-#endif /* VMS */
+		    doshell();
 		    break;
-		    }
 
 		/*
 		 * Range commands:
 		 */
 
-		case '/':
+		case 'r':
 		    error(
 "Range: x:erase v:value c:copy f:fill d:def l:lock U:unlock S:show u:undef F:fmt");
 		    if (braille) move(1, 0);
@@ -1178,38 +1169,38 @@ main (int argc, char  **argv)
 		    error("");
 		    switch (c) {
 		    case 'l':
-			(void) sprintf(line,"lock [range] ");
+			(void) sprintf(line, "lock [range] ");
 			linelim = strlen(line);
 			insert_mode();
 			startshow();
 			break;
 		    case 'U':
-			(void) sprintf(line,"unlock [range] ");
+			(void) sprintf(line, "unlock [range] ");
 			linelim = strlen(line);
 			insert_mode();
 			startshow();
 			break;
 		    case 'c':
-			(void) sprintf(line,"copy [dest_range src_range] ");
+			(void) sprintf(line, "copy [dest_range src_range] ");
 			linelim = strlen(line);
 			insert_mode();
 			startshow();
 			break;
 		    case 'm':
-			(void) sprintf(line,"move [destination src_range] %s ",
+			(void) sprintf(line, "move [destination src_range] %s ",
 				v_name(currow, curcol));
 			linelim = strlen(line);
 			insert_mode();
 		        write_line(ctl('v'));
 			break;
 		    case 'x':
-			(void) sprintf(line,"erase [range] ");
+			(void) sprintf(line, "erase [range] ");
 			linelim = strlen(line);
 			insert_mode();
 			startshow();
 			break;
 		    case 'y':
-			(void) sprintf(line,"yank [range] ");
+			(void) sprintf(line, "yank [range] ");
 			linelim = strlen(line);
 			insert_mode();
 			startshow();
@@ -1221,18 +1212,18 @@ main (int argc, char  **argv)
 			startshow();
 			break;
 		    case 'f':
-			(void) sprintf(line,"fill [range start inc] ");
+			(void) sprintf(line, "fill [range start inc] ");
 			linelim = strlen(line);
 			insert_mode();
 			startshow();
 			break;
 		    case 'd':
-			(void) sprintf(line,"define [string range] \"");
+			(void) sprintf(line, "define [string range] \"");
 			linelim = strlen(line);
 			insert_mode();
 			break;
 		    case 'u':
-			(void) sprintf(line,"undefine [range] ");
+			(void) sprintf(line, "undefine [range] ");
 			linelim = strlen(line);
 			insert_mode();
 			break;
@@ -1245,22 +1236,22 @@ main (int argc, char  **argv)
 			error("");
 			switch (c) {
 			    case 't':
-				sprintf(line,"frametop [<outrange> rows] ");
+				sprintf(line, "frametop [<outrange> rows] ");
 				break;
 			    case 'b':
-				sprintf(line,"framebottom [<outrange> rows] ");
+				sprintf(line, "framebottom [<outrange> rows] ");
 				break;
 			    case 'l':
-				sprintf(line,"frameleft [<outrange> cols] ");
+				sprintf(line, "frameleft [<outrange> cols] ");
 				break;
 			    case 'r':
-				sprintf(line,"frameright [<outrange> cols] ");
+				sprintf(line, "frameright [<outrange> cols] ");
 				break;
 			    case 'a':
-				sprintf(line,"frame [<outrange> inrange] ");
+				sprintf(line, "frame [<outrange> inrange] ");
 				break;
 			    case 'u':
-				sprintf(line,"unframe [<range>] ");
+				sprintf(line, "unframe [<range>] ");
 				break;
 			    case ESC:
 			    case ctl('g'):
@@ -1285,12 +1276,15 @@ main (int argc, char  **argv)
 			startshow();
 			break;
 		    case 'C':
-			(void) sprintf(line,"color [range color#] ");
+			(void) sprintf(line, "color [range color#] ");
 			linelim = strlen(line);
 			insert_mode();
 			startshow();
 			break;
-		    case 'S':	/* Show color definitions and ranges */
+		    case 'S':
+			/* Show color definitions and various types of
+			 * ranges
+			 */
 			if (!are_ranges() && !are_frames() && !are_colors())
 			    error("Nothing to show");
 			else {
@@ -1325,6 +1319,24 @@ main (int argc, char  **argv)
 			insert_mode();
 			startshow();
 			break;
+		    case '{':
+			(void) sprintf(line, "leftjustify [range] ");
+			linelim = strlen(line);
+			insert_mode();
+			startshow();
+			break;
+		    case '}':
+			(void) sprintf(line, "rightjustify [range] ");
+			linelim = strlen(line);
+			insert_mode();
+			startshow();
+			break;
+		    case '|':
+			(void) sprintf(line, "center [range] ");
+			linelim = strlen(line);
+			insert_mode();
+			startshow();
+			break;
 		    case ESC:
 		    case ctl('g'):
 			break;
@@ -1334,8 +1346,14 @@ main (int argc, char  **argv)
 		    }
 		    break;
 
+		case '~':
+		    (void) sprintf(line, "abbrev \"");
+		    linelim = strlen(line);
+		    insert_mode();
+		    break;
+
 		case '"':
-		    error("Select buffer (0-9 or a-z):");
+		    error("Select buffer (a-z or 0-9):");
 		    if ((c=nmgetch()) == ESC || c == ctl('g')) {
 			error("");
 		    } else if (c >= '0' && c <= '9') {
@@ -1406,6 +1424,13 @@ main (int argc, char  **argv)
 				break;
 
 			    case 'p':
+				if (rcqual == '.') {
+				    (void) sprintf(line, "pullcopy ");
+				    linelim = strlen(line);
+				    insert_mode();
+				    startshow();
+				    break;
+				}
 				while (arg--)		pullcells(rcqual);
 				break;
 
@@ -1427,7 +1452,7 @@ main (int argc, char  **argv)
 				} else
 				    valueize_area(0, curcol,
 					    maxrow, curcol + arg - 1);
-				modflg = 1;
+				modflg++;
 				break;
 
 			    case 'Z':
@@ -1457,31 +1482,11 @@ main (int argc, char  **argv)
 		    }
 
 		case '$':
-		    {
-		    register struct ent *p;
-
-		    remember(0);
-		    curcol = maxcols - 1;
-		    while (!VALID_CELL(p, currow, curcol) && curcol > 0)
-			curcol--;
-		    rowsinrange = 1;
-		    colsinrange = fwidth[curcol];
-		    remember(1);
+		    rightlimit();
 		    break;
-		    }
 		case '#':
-		    {
-		    register struct ent *p;
-
-		    remember(0);
-		    currow = maxrows - 1;
-		    while (!VALID_CELL(p, currow, curcol) && currow > 0)
-			currow--;
-		    rowsinrange = 1;
-		    colsinrange = fwidth[curcol];
-		    remember(1);
+		    gotobottom();
 		    break;
-		    }
 		case 'w':
 		    {
 		    register struct ent *p;
@@ -1534,11 +1539,7 @@ main (int argc, char  **argv)
 		    break;
 		    }
 		case '^':
-		    remember(0);
-		    currow = 0;
-		    rowsinrange = 1;
-		    colsinrange = fwidth[curcol];
-		    remember(1);
+		    gototop();
 		    break;
 #ifdef KEY_HELP
 		case KEY_HELP:
@@ -1592,37 +1593,31 @@ main (int argc, char  **argv)
 		    break;
 		case '{':
 		    {
-			struct ent *p = lookat(currow, curcol);
-			if (p->label) {
-			    p->flags &= ~is_label;
-			    p->flags |= is_leftflush | is_changed;
-			    changed++;
-			    modflg++;
-			} else
+			struct ent *p = *ATBL(tbl, currow, curcol);
+
+			if (p && p->label)
+			    ljustify(currow, curcol, currow, curcol);
+			else
 			    error("Nothing to justify");
 			break;
 		    }
 		case '}':
 		    {
-			struct ent *p = lookat(currow, curcol);
-			if (p->label) {
-			    p->flags &= ~(is_label | is_leftflush);
-			    p->flags |= is_changed;
-			    changed++;
-			    modflg++;
-			} else
+			struct ent *p = *ATBL(tbl, currow, curcol);
+
+			if (p && p->label)
+			    rjustify(currow, curcol, currow, curcol);
+			else
 			    error("Nothing to justify");
 			break;
 		    }
 		case '|':
 		    {
-			struct ent *p = lookat(currow, curcol);
-			if (p->label) {
-			    p->flags &= ~is_leftflush;
-			    p->flags |= is_label | is_changed;
-			    changed++;
-			    modflg++;
-			} else
+			struct ent *p = *ATBL(tbl, currow, curcol);
+
+			if (p && p->label)
+			    center(currow, curcol, currow, curcol);
+			else
 			    error("Nothing to center");
 			break;
 		    }
@@ -1708,6 +1703,9 @@ main (int argc, char  **argv)
 		    linelim = strlen(line);
 		    insert_mode();
 		    break;
+		case 'n':
+		    go_last();
+		    break;
 		case 'P':
 		    (void) sprintf(line, "put [\"dest\" range] \"");
 
@@ -1814,13 +1812,13 @@ main (int argc, char  **argv)
 		    insert_mode();
 		    break;
 		case 'S':	/* set options */
-		    (void) sprintf (line, "set ");
+		    (void) sprintf(line, "set ");
 		    error("Options:byrows,bycols,iterations=n,tblstyle=(0|tbl|latex|slatex|tex|frame),<MORE>");
 		    linelim = strlen(line);
 		    insert_mode();
 		    break;
 		case 'T':	/* tbl output */
-		    (void) sprintf (line, "tbl [\"dest\" range] \"");
+		    (void) sprintf(line, "tbl [\"dest\" range] \"");
 
 /* See the comments under "case 'W':" above for an explanation of the
  * logic here.
@@ -1919,13 +1917,7 @@ main (int argc, char  **argv)
 		    break;
 #ifdef KEY_HOME
 		case KEY_HOME:
-		    remember(0);
-		    currow = 0;
-		    curcol = 0;
-		    rowsinrange = 1;
-		    colsinrange = fwidth[curcol];
-		    remember(1);
-		    FullUpdate++;
+		    gohome();
 		    break;
 #endif
 		case 'L':
@@ -1941,10 +1933,18 @@ main (int argc, char  **argv)
 		    break;
 		case 'c':
 		    error("Copy marked cell:");
-		    if ((c=nmgetch()) == ESC || c == ctl('g')) {
+		    if ((c = nmgetch()) == ESC || c == ctl('g')) {
 			error("");
 			break;
 		    }
+		    if (c == '.') {
+			copy(NULL, NULL, lookat(currow, curcol), NULL);
+			(void) sprintf(line, "copy [dest_range src_range] ");
+			linelim = strlen(line);
+			insert_mode();
+			startshow();
+			break;
+		    }
 		    if (c == '`' || c == '\'')
 			c = 0;
 		    else if (!(((c -= ('a' - 1)) > 0 && c < 27) ||
@@ -1958,30 +1958,40 @@ main (int argc, char  **argv)
 		    }
 		    error("");
 		    {
-			register struct ent *p = *ATBL(tbl, savedrow[c], savedcol[c]);
+			struct ent *p = *ATBL(tbl, savedrow[c], savedcol[c]);
 			int c1;
-			register struct ent *n;
-			if (!p)
-			    break;
-			FullUpdate++;
-			modflg++;
+			struct ent *n;
+
 			for (c1 = curcol; arg-- && c1 < maxcols; c1++) {
-			    n = lookat(currow, c1);
-			    (void) clearent(n);
+			    if ((n = *ATBL(tbl, currow, c1))) {
+				if (n->flags & is_locked)
+				    continue;
+				if (!p) {
+				    (void) clearent(n);
+				    continue;
+				}
+			    } else {
+				if (!p) break;
+				n = lookat(currow, c1);
+			    }
 			    copyent(n, p, currow - savedrow[c],
 				    c1 - savedcol[c], 0, 0, maxrow, maxcol, 0);
+			    n->flags |= is_changed;
 			}
-		    break;
+
+			FullUpdate++;
+			modflg++;
+			break;
 		    }
 		case '`':
 		case '\'':
 			dotick(c);
 			break;
-		case 'n':
+		case '*':
 		    {
 			register struct ent *p;
 
-			error("Note: Add/Delete/Show?");
+			error("Note: Add/Delete/Show/*(go to note)?");
 			if ((c = nmgetch()) == ESC || c == ctl('g')) {
 			    error("");
 			    break;
@@ -1993,12 +2003,16 @@ main (int argc, char  **argv)
 			    insert_mode();
 			    write_line(ctl('v'));
 			    error("");
+			    FullUpdate++;
 			    break;
 			}
 			if (c == 'd' || c == 'D') {
 			    p = lookat(currow, curcol);
 			    p->nrow = p->ncol = -1;
+			    p->flags |= is_changed;
 			    error("");
+			    modflg++;
+			    FullUpdate++;
 			    break;
 			}
 			if (c == 's' || c == 'S') {
@@ -2008,12 +2022,14 @@ main (int argc, char  **argv)
 			    error("Highlighted cells have attached notes.");
 			    break;
 			}
+			if (c == '*') {
+			    gotonote();
+			    error("");
+			    break;
+			}
 			error("Invalid command");
 			break;
 		    }
-		case '*':
-			gotonote();
-			break;
 		case 'z':
 		    switch (c = nmgetch()) {
 			case ctl('m'):
@@ -2045,11 +2061,13 @@ main (int argc, char  **argv)
 			    break;
 		    }
 		    break;
+#ifdef KEY_RESIZE
 		case KEY_RESIZE:
 #ifndef	SIGWINCH
 		    winchg();
 #endif
 		    break;
+#endif
 		default:
 		    if ((toascii(c)) != c)
 			error ("Weird character, decimal %d\n",
@@ -2064,6 +2082,7 @@ main (int argc, char  **argv)
     inloop = modcheck(" before exiting");
     }				/*  while (inloop) */
     stopdisp();
+    write_hist();
 #ifdef VMS	/* Until VMS "fixes" exit we should say 1 here */
     exit (1);
 #else
@@ -2112,7 +2131,9 @@ signals()
     (void) signal(SIGQUIT, dump_me);
     (void) signal(SIGPIPE, nopipe);
     (void) signal(SIGALRM, time_out);
+#ifndef __DJGPP__
     (void) signal(SIGBUS, doquit);
+#endif
 #endif
     (void) signal(SIGTERM, doquit);
     (void) signal(SIGFPE, doquit);
@@ -2166,6 +2187,7 @@ doquit()
 	diesave();
 	stopdisp();
     }
+    write_hist();
     exit (1);
 }
 
