@@ -6,7 +6,7 @@
  *			University of Maryland
  *		R. Bond  12/86
  *		More mods by Alan Silverstein, 3-4/88, see list of changes.
- *		$Revision: 7.15 $
+ *		$Revision: 7.16 $
  *
  */
 
@@ -35,11 +35,11 @@
 #define DEFPREC   2
 #define DEFREFMT  REFMTFIX /* Make default format fixed point  THA 10/14/90 */
 
-#define FKEYS		24	/* Number of function keys available */
-#define HISTLEN		10	/* Number of history entries for vi emulation */
-#define CPAIRS		 8	/* Number of color pairs available */
-#define COLFORMATS	10	/* Number of custom column formats */
-#define DELBUFSIZE	40	/* Number of named buffers + 4 */
+#define FKEYS		 24	/* Number of function keys available */
+#define HISTLEN		100	/* Number of history entries for vi emulation */
+#define CPAIRS		  8	/* Number of color pairs available */
+#define COLFORMATS	 10	/* Number of custom column formats */
+#define DELBUFSIZE	 40	/* Number of named buffers + 4 */
 #ifdef PSC
 # define error(msg)	fprintf(stderr, msg);
 #else
@@ -48,6 +48,13 @@
 #define	FBUFLEN	1024	/* buffer size for a single field */
 #define	PATHLEN	1024	/* maximum path length */
 
+#ifndef DFLT_PAGER
+#define	DFLT_PAGER "more"	/* more is probably more widespread than less */
+#endif /* DFLT_PAGER */
+
+#define MAXCMD 160	/* for ! command and commands that use the pager */
+
+
 #ifndef A_CHARTEXT	/* Should be defined in curses.h */
 #define A_CHARTEXT 0xff
 #endif
@@ -127,37 +134,9 @@ struct ent {
     char cellerror;		/* error in a cell? */
 };
 
-struct colorpair {
-    int fg;
-    int bg;
-    struct enode *expr;
-};
-
-/* stores a range (left, right) */
-struct range {
-    struct ent_ptr r_left, r_right;
-    char *r_name;			/* possible name for this range */
-    struct range *r_next, *r_prev;	/* chained ranges */
-    int r_is_range;
-};
-
 #define FIX_ROW 1
 #define FIX_COL 2
 
-/* stores a framed range (left, right) */
-struct frange {
-    struct ent *or_left, *or_right;	/* outer range */
-    struct ent *ir_left, *ir_right;	/* inner range */
-    struct frange *r_next, *r_prev;	/* chained ranges */
-};
-
-/* stores a color range (left, right) */
-struct crange {
-    struct ent *r_left, *r_right;
-    int r_color;
-    struct crange *r_next, *r_prev;	/* chained ranges */
-};
-
 /* stores type of operation this cell will perform */
 struct enode {
     int op;
@@ -174,6 +153,48 @@ struct enode {
     } e;
 };
 
+/* stores a range (left, right) */
+struct range {
+    struct ent_ptr r_left, r_right;
+    char *r_name;			/* possible name for this range */
+    struct range *r_next, *r_prev;	/* chained ranges */
+    int r_is_range;
+};
+
+/* stores a framed range (left, right) */
+struct frange {
+    struct ent *or_left, *or_right;	/* outer range */
+    struct ent *ir_left, *ir_right;	/* inner range */
+    struct frange *r_next, *r_prev;	/* chained ranges */
+};
+
+/* stores a color range (left, right) */
+struct crange {
+    struct ent *r_left, *r_right;
+    int r_color;
+    struct crange *r_next, *r_prev;	/* chained ranges */
+};
+
+struct colorpair {
+    int fg;
+    int bg;
+    struct enode *expr;
+};
+
+/* stores an abbreviation and its expansion */
+struct abbrev {
+    char *abbr;
+    char *exp;
+    struct abbrev *a_next, *a_prev;
+};
+
+struct impexfilt {
+    char ext[PATHLEN];
+    char plugin[PATHLEN];
+    char type;
+    struct impexfilt *next;
+};
+
 /* Use this structure to save the last 'g' command */
 struct go_save {
     int		g_type;
@@ -214,58 +235,66 @@ struct go_save {
 #define TAN (OP_BASE + 14)
 #define DTR (OP_BASE + 15)
 #define RTD (OP_BASE + 16)
-#define MIN (OP_BASE + 17)
-#define MAX (OP_BASE + 18)
-#define RND (OP_BASE + 19)
-#define HOUR (OP_BASE + 20)
-#define MINUTE (OP_BASE + 21)
-#define SECOND (OP_BASE + 22)
-#define MONTH (OP_BASE + 23)
-#define DAY (OP_BASE + 24)
-#define YEAR (OP_BASE + 25)
-#define NOW (OP_BASE + 26)
-#define DATE (OP_BASE + 27)
-#define FMT (OP_BASE + 28)
-#define SUBSTR (OP_BASE + 29)
-#define STON (OP_BASE + 30)
-#define EQS (OP_BASE + 31)
-#define EXT (OP_BASE + 32)
-#define ELIST (OP_BASE + 33	/* List of expressions */)
-#define LMAX  (OP_BASE + 34)
-#define LMIN  (OP_BASE + 35)
-#define NVAL (OP_BASE + 36)
-#define SVAL (OP_BASE + 37)
-#define PV (OP_BASE + 38)
-#define FV (OP_BASE + 39)
-#define PMT (OP_BASE + 40)
-#define STINDEX (OP_BASE + 41)
-#define LOOKUP (OP_BASE + 42)
-#define ATAN2 (OP_BASE + 43)
-#define INDEX (OP_BASE + 44)
-#define DTS (OP_BASE + 45)
-#define TTS (OP_BASE + 46)
-#define ABS (OP_BASE + 47)
-#define HLOOKUP (OP_BASE + 48)
-#define VLOOKUP (OP_BASE + 49)
-#define ROUND (OP_BASE + 50)
-#define IF (OP_BASE + 51)
-#define MYROW (OP_BASE + 52)
-#define MYCOL (OP_BASE + 53)
-#define COLTOA (OP_BASE + 54)
-#define UPPER (OP_BASE + 55)
-#define LOWER (OP_BASE + 56)
-#define CAPITAL (OP_BASE + 57)
-#define NUMITER (OP_BASE + 58)
-#define ERR_ (OP_BASE + 59)
-#define PI_ (OP_BASE + 60)
-#define BLACK (OP_BASE + 61)
-#define RED (OP_BASE + 62)
-#define GREEN (OP_BASE + 63)
-#define YELLOW (OP_BASE + 64)
-#define BLUE (OP_BASE + 65)
-#define MAGENTA (OP_BASE + 66)
-#define CYAN (OP_BASE + 67)
-#define WHITE (OP_BASE + 68)
+#define SUM (OP_BASE + 17)
+#define PROD (OP_BASE + 18)
+#define AVG (OP_BASE + 19)
+#define COUNT (OP_BASE + 20)
+#define STDDEV (OP_BASE + 21)
+#define MAX (OP_BASE + 22)
+#define MIN (OP_BASE + 23)
+#define RND (OP_BASE + 24)
+#define HOUR (OP_BASE + 25)
+#define MINUTE (OP_BASE + 26)
+#define SECOND (OP_BASE + 27)
+#define MONTH (OP_BASE + 28)
+#define DAY (OP_BASE + 29)
+#define YEAR (OP_BASE + 30)
+#define NOW (OP_BASE + 31)
+#define DATE (OP_BASE + 32)
+#define FMT (OP_BASE + 33)
+#define SUBSTR (OP_BASE + 34)
+#define STON (OP_BASE + 35)
+#define EQS (OP_BASE + 36)
+#define EXT (OP_BASE + 37)
+#define ELIST (OP_BASE + 38)	/* List of expressions */
+#define LMAX  (OP_BASE + 39)
+#define LMIN  (OP_BASE + 40)
+#define NVAL (OP_BASE + 41)
+#define SVAL (OP_BASE + 42)
+#define PV (OP_BASE + 43)
+#define FV (OP_BASE + 44)
+#define PMT (OP_BASE + 45)
+#define STINDEX (OP_BASE + 46)
+#define LOOKUP (OP_BASE + 47)
+#define ATAN2 (OP_BASE + 48)
+#define INDEX (OP_BASE + 49)
+#define DTS (OP_BASE + 50)
+#define TTS (OP_BASE + 51)
+#define ABS (OP_BASE + 52)
+#define HLOOKUP (OP_BASE + 53)
+#define VLOOKUP (OP_BASE + 54)
+#define ROUND (OP_BASE + 55)
+#define IF (OP_BASE + 56)
+#define FILENAME (OP_BASE + 57)
+#define MYROW (OP_BASE + 58)
+#define MYCOL (OP_BASE + 59)
+#define LASTROW (OP_BASE + 60)
+#define LASTCOL (OP_BASE + 61)
+#define COLTOA (OP_BASE + 62)
+#define UPPER (OP_BASE + 63)
+#define LOWER (OP_BASE + 64)
+#define CAPITAL (OP_BASE + 65)
+#define NUMITER (OP_BASE + 66)
+#define ERR_ (OP_BASE + 67)
+#define PI_ (OP_BASE + 68)
+#define BLACK (OP_BASE + 69)
+#define RED (OP_BASE + 70)
+#define GREEN (OP_BASE + 71)
+#define YELLOW (OP_BASE + 72)
+#define BLUE (OP_BASE + 73)
+#define MAGENTA (OP_BASE + 74)
+#define CYAN (OP_BASE + 75)
+#define WHITE (OP_BASE + 76)
 
 /* flag values */
 #define is_valid     0001
@@ -291,6 +320,10 @@ struct go_save {
 #define BYCOLS 1
 #define BYROWS 2
 
+/* values for showrange for ranges of rows or columns */
+#define SHOWROWS 2
+#define SHOWCOLS 3
+
 /* tblprint style output for: */
 #define	TBL	1		/* 'tbl' */
 #define	LATEX	2		/* 'LaTeX' */
@@ -333,8 +366,9 @@ extern	char line[FBUFLEN];
 extern	int linelim;
 extern	int changed;
 extern	struct ent *delbuf[DELBUFSIZE];
+extern	char *delbuffmt[DELBUFSIZE];
 extern	int dbidx;
-extern	int qbuf;	/* buffer no specified by " command */
+extern	int qbuf;		/* buffer no. specified by `"' command */
 extern	int showsc, showsr;
 extern	int showrange;		/* Causes ranges to be highlighted */
 extern	int cellassign;
@@ -342,15 +376,17 @@ extern	int macrofd;
 extern	int cslop;
 extern	int usecurses;
 extern	int brokenpipe;		/* Set to true if SIGPIPE is received */
+extern  char dpoint;	/* country-dependent decimal point from locale */
+extern  char thsep;	/* country-dependent thousands separator from locale */
 
 extern	FILE *openfile(char *fname, int *rpid, int *rfd);
 extern	char *coltoa(int col);
+extern	char *findplugin(char *ext, char type);
 extern	char *findhome(char *path);
 extern	char *r_name(int r1, int c1, int r2, int c2);
 extern	char *scxmalloc(unsigned n);
 extern	char *scxrealloc(char *ptr, unsigned n);
 extern	char *seval(register struct enode *se);
-extern	char *strrchr();
 extern	char *v_name(int row, int col);
 extern	double eval(register struct enode *e);
 extern	int any_locked_cells(int r1, int c1, int r2, int c2);
@@ -358,6 +394,7 @@ extern	int are_colors();
 extern	int are_frames();
 extern	int are_ranges();
 extern	int atocol(char *string, int len);
+extern	int creadfile(char *save, int  eraseflg);
 extern	int cwritefile(char *fname, int r0, int c0, int rn, int cn);
 extern	int engformat(int fmt, int width, int lprecision, double val,
 	char *buf, int buflen);
@@ -371,8 +408,11 @@ extern	int growtbl(int rowcol, int toprow, int topcol);
 extern	int locked_cell(int r, int c);
 extern	int modcheck(char *endstr);
 extern	int nmgetch();
+extern	int plugin_exists(char *name, int len, char *path);
+extern	int readfile(char *fname, int eraseflg);
 extern	int writefile(char *fname, int r0, int c0, int rn, int cn);
 extern	int yn_ask(char *msg);
+extern	struct abbrev *find_abbr(char *abbrev, int len, struct abbrev **prev);
 extern	struct colorpair *cpairs[8];
 extern	struct enode *copye(register struct enode *e, int Rdelta, int Cdelta,
 	int r1, int c1, int r2, int c2, int transpose);
@@ -391,6 +431,7 @@ extern	void add_frange(struct ent *or_left, struct ent *or_right,
 	int leftcols, int rightcols);
 extern	void add_range(char *name, struct ent_ptr left, struct ent_ptr right,
 	int is_range);
+extern	void addplugin(char *ext, char *plugin, char type);
 extern	void backcol(int arg);
 extern	void backrow(int arg);
 extern	void change_color(int pair, struct enode *e);
@@ -407,10 +448,10 @@ extern	void copy(struct ent *dv1, struct ent *dv2, struct ent *v1,
 	struct ent *v2);
 extern	void copyent(register struct ent *n, register struct ent *p,
 	int dr, int dc, int r1, int c1, int r2, int c2, int transpose);
-extern	void creadfile(char *save, int  eraseflg);
 extern	void decompile(register struct enode *e, int priority);
 extern	void deleterow(register int arg);
 extern	void del_range(struct ent *left, struct ent *right);
+extern	void del_abbr(char *abbrev);
 extern	void deraw(int ClearLastLine);
 extern	void diesave();
 extern	void doend(int rowinc, int colinc);
@@ -426,7 +467,7 @@ extern	void edit_mode();
 extern	void edits(int row, int col);
 extern	void editv(int row, int col);
 extern	void efree(struct enode *e);
-extern	void erase_area(int sr, int sc, int er, int ec);
+extern	void erase_area(int sr, int sc, int er, int ec, int ignorelock);
 extern	void erasedb();
 extern	void eraser(struct ent *v1, struct ent *v2);
 extern	void fgetnum(int r0, int c0, int rn, int cn, int fd);
@@ -442,7 +483,7 @@ extern	void formatcol(int arg);
 extern	void format_cell(struct ent *v1, struct ent *v2, char *s);
 extern	void forwcol(int arg);
 extern	void forwrow(int arg);
-extern	void free_ent(register struct ent *p);
+extern	void free_ent(register struct ent *p, int unlock);
 extern	void getexp(int r0, int c0, int rn, int cn, int fd);
 extern	void getfmt(int r0, int c0, int rn, int cn, int fd);
 extern	void getformat(int col, int fd);
@@ -479,7 +520,7 @@ extern	void num_search(double n, int firstrow, int firstcol, int lastrow,
 extern	void printfile(char *fname, int r0, int c0, int rn, int cn);
 extern	void pullcells(int to_insert);
 extern	void query(char *s, char *data);
-extern	void readfile(char *fname, int eraseflg);
+extern	void read_hist();
 extern	void remember(int save);
 extern	void resetkbd();
 extern	void rowshow_op();
@@ -515,6 +556,7 @@ extern	void write_colors(FILE *f, int indent);
 extern	void write_cranges(FILE *f);
 extern	void write_fd(register FILE *f, int r0, int c0, int rn, int cn);
 extern	void write_franges(FILE *f);
+extern	void write_hist();
 extern	void write_line(int c);
 extern	void write_ranges(FILE *f);
 extern	void yank_area(int sr, int sc, int er, int ec);
@@ -549,6 +591,7 @@ extern	int calc_order;
 extern	int autocalc;
 extern	int autolabel;
 extern	int autoinsert;
+extern	int autowrap;
 extern	int optimize;
 extern	int numeric;
 extern	int showcell;
