@@ -7,7 +7,7 @@
  *
  *              More mods Robert Bond, 12/86
  *		More mods by Alan Silverstein, 3-4/88, see list of changes.
- *		$Revision: 7.15 $
+ *		$Revision: 7.16 $
  *
  */
 
@@ -33,13 +33,11 @@ extern int VMS_read_raw;   /*sigh*/
 #define nonl()	 (_tty.sg_flags &= ~CRMOD, _pfast = TRUE, stty(_tty_ch, &_tty))
 #endif
 
-#define SHOWROWS 2
-#define SHOWCOLS 3
-
 void	repaint(int x, int y, int len, int attron, int attroff);
 
 char	under_cursor = ' '; /* Data under the < cursor */
 char	mode_ind = 'i';
+char	search_ind = ' ';
 extern	char    revmsg[];
 
 int	lines, cols;
@@ -611,6 +609,8 @@ update(int anychanged)		/* did any cell really change in value? */
 	    k = (fwidth[i] - strlen(coltoa(i)))/2;
 	    if (fwidth[i] == 1)
 		(void) printw("%1s", coltoa(i%26));
+	    else if (braille)
+	        (void) printw("%-*s", fwidth[i], coltoa(i));
 	    else
 	        (void) printw("%*s%-*s", k, "", fwidth[i]-k, coltoa(i));
 	    col += fwidth[i];
@@ -783,6 +783,7 @@ update(int anychanged)		/* did any cell really change in value? */
 			char *cfmt;
 			int note;
 
+			*field = '\0';
 			note = (*pp)->nrow >= 0 ? 1 : 0;
 			cfmt = (*pp)->format ? (*pp)->format :
 			    (realfmt[col] >= 0 && realfmt[col] < COLFORMATS &&
@@ -804,8 +805,8 @@ update(int anychanged)		/* did any cell really change in value? */
 					field, sizeof(field));
 			} else {
 			    (void) engformat(realfmt[col], fwidth[col] - note,
-				precision[col], (*pp)->v, 
-				field, sizeof(field));
+				    precision[col], (*pp)->v, 
+				    field, sizeof(field));
 			}
 			if (strlen(field) > fwidth[col]) {
 			    for (i = 0; i < fwidth[col]; i++) {
@@ -927,7 +928,8 @@ update(int anychanged)		/* did any cell really change in value? */
 	    if ((unsigned char) line[i] < ' ')
 		ctlchars++;
 	(void) addch(mode_ind);
-	(void) addstr("> ");
+	(void) addch('>');
+	(void) addch(search_ind);
 	(void) addstr(line);
 	if (!braille || (!message && mode_ind != 'v'))
 	    (void) move((linelim+3+ctlchars)/cols, (linelim+3+ctlchars)%cols);
@@ -936,7 +938,7 @@ update(int anychanged)		/* did any cell really change in value? */
 	else if (braillealt)
 	    move(1, 16);
 	else
-	    move(lastmy,lastmx+(fwidth[lastcol]-strlen(coltoa(lastcol)))/2);
+	    move(lastmy, lastmx);
     } else {
 	if (showtop) {			/* show top line */
 	    register struct ent *p1;
@@ -994,7 +996,7 @@ update(int anychanged)		/* did any cell really change in value? */
 		if (p1->flags & is_valid) {
 		    /* has value or num expr */
 		    if ((!(p1->expr)) || (p1->flags & is_strexpr))
-			(void) sprintf (line, "%.15g", p1->v);
+			(void) sprintf(line, "%.15g", p1->v);
 
 		    (void) addch('[');
 		    (void) addstr(line);
@@ -1015,7 +1017,28 @@ update(int anychanged)		/* did any cell really change in value? */
 	    else if (braillealt)
 		move(0, 0);
 	    else
-		move(lastmy,lastmx+(fwidth[lastcol]-strlen(coltoa(lastcol)))/2);
+		move(lastmy, lastmx);
+	else if (showcell)
+	    move(lines - 1, cols - 1);
+	else
+	    (void) move(lastmy, lastmx+fwidth[lastcol]);
+    }
+
+    if (color && has_colors())
+	color_set(1, NULL);
+
+    if (revmsg[0]) {
+	(void) move(0, 0);
+	(void) clrtoeol();	/* get rid of topline display */
+	(void) printw(revmsg);
+	*revmsg = '\0';		/* don't show it again */
+	if (braille)
+	    if (message)
+		move(1, 0);
+	    else if (braillealt)
+		move(0, 0);
+	    else
+		move(lastmy, lastmx);
 	else if (showcell)
 	    move(lines - 1, cols - 1);
 	else
@@ -1034,7 +1057,7 @@ update(int anychanged)		/* did any cell really change in value? */
 	    if (message)
 		move(1, 0);
 	    else
-		move(lastmy,lastmx+(fwidth[lastcol]-strlen(coltoa(lastcol)))/2);
+		move(lastmy, lastmx);
 	else if (showcell)
 	    move(lines - 1, cols - 1);
 	else
@@ -1064,7 +1087,7 @@ yyerror(char *err)
     if (usecurses) {
 	if (seenerr) return;
 	seenerr++;
-	(void) move(1,0);
+	(void) move(1, 0);
 	(void) clrtoeol();
 	(void) printw("%s: %.*s<=%s", err, linelim, line, line + linelim);
     } else
@@ -1214,7 +1237,6 @@ goraw()
 	kbd_again();
 	if (color && has_colors())
 	    bkgdset(COLOR_PAIR(1) | ' ');
-	(void) clear();
 	FullUpdate++;
     }
 }
