@@ -1,7 +1,7 @@
 /*	SC	A Spreadsheet Calculator
  *
  *	One line vi emulation
- *	$Revision: 7.15 $
+ *	$Revision: 7.16 $
  */
 
 #include <sys/types.h>
@@ -16,38 +16,65 @@
 #include <signal.h>
 #include <curses.h>
 #include <ctype.h>
+#include <stdlib.h>
 #include "sc.h"
 
+#if defined(REGCOMP)
+#include <regex.h>
+#endif
+#if defined(RE_COMP)
+extern char *re_comp(char *s);
+extern char *re_exec(char *s);
+#endif
+#if defined(REGCMP)
+char *regcmp();
+char *regex();
+#endif
+
+void doshell();
+void gohome();
+void leftlimit();
+void rightlimit();
+void gototop();
+void gotobottom();
+
 #define istext(a) (isalnum(a) || ((a) == '_'))
 
+#define bool	int
+#define true	1
+#define false	0
+
 static void append_line();
 static void back_hist();
 static int  back_line(int arg);
-static int  back_word(int arg);
+static int  back_word(int arg, int big_word);
 static void back_space();
+static void change_case(int arg);
 static void col_0();
 static void cr_line();
 static void del_in_line(int arg, int back_null);
 static void del_to_end();
+static void doabbrev();
 static void dogoto();
 static void dotab();
 static void dotcmd();
 static int  find_char(int arg, int n);
 static void for_hist();
 static int  for_line(int arg, int stop_null);
-static int  for_word(int arg, int end_word, int stop_null);
+static int  for_word(int arg, int end_word, int big_word, int stop_null);
+static int  istart;
 static void last_col();
 static void match_paren();
-static void change_case(int arg);
+static void readhistfile(FILE *fp);
 static void rep_char();
 static void replace_in_line(int c);
 static void replace_mode();
 static void restore_it();
 static void savedot(int c);
 static void save_hist();
-static void search_again();
+static void search_again(bool reverse);
 static void search_hist();
-static void search_mode();
+static void search_mode(char sind);
 static void stop_edit();
 static int  to_char(int arg, int n);
 static void u_save(int c);
@@ -56,8 +83,10 @@ static void yank_chars(register int first, register int last, int delete);
 
 extern int framerows;		/* Rows in current frame */
 extern char mode_ind;		/* Mode indicator */
+extern char search_ind;		/* Search indicator */
 extern int lcols;		/* Spreadsheet Column the cursor was in last */
-char *completethis = NULL;
+char	*completethis = NULL;
+int	search_dir;		/* Search direction:  forward = 0; back = 1 */
 
 /* values for mode below */
 
@@ -79,7 +108,18 @@ struct	hist {
 static int	histp = 0;
 static int	lasthist = 0;
 static int	endhist = -1;
+static int	histsessionstart = 0;
+static int	histsessionnew = 0;
+
+#ifdef REGCOMP
+static regex_t	preg;
+static regex_t	*last_search = NULL;
+static int	errcode;
+#else
+#ifndef RE_COMP
 static char	*last_search = NULL;
+#endif
+#endif
 static char	*undo_line = NULL;
 static int	undo_lim;
 static char	dotb[DOTLEN];
@@ -88,7 +128,7 @@ static int	do_dot = 0;
 static int	nosavedot = 1;
 static int	dotarg = 1;
 static char	putbuf[FBUFLEN];
-static int	findfunc = '0';
+static int	findfunc = '\0';
 static int	findchar = 1;
 static int	finddir = 0;
 
@@ -106,7 +146,10 @@ write_line(int c)
 	case KEY_BACKSPACE:
 	case (ctl('h')):	linelim = back_line(arg);		break;
 	case (ctl('i')):	dotab();				break;
-	case (ctl('m')):	cr_line();				break;
+	case (ctl('m')):	if (search_ind != ' ')
+				    search_hist();
+				else
+				    cr_line();				break;
 	case 'v':
 	case (ctl('v')):	toggle_navigate_mode();			break;
 	case ESC:	stop_edit();					break;
@@ -116,6 +159,7 @@ write_line(int c)
 	case (ctl('e')):
 	case '$':	last_col();					break;
 	case '.':	dotcmd();					break;
+	case '!':	doshell();					break;
 	case ';':	if (findfunc)
 			    ungetch(findchar);
 			else
@@ -141,28 +185,33 @@ write_line(int c)
 #ifdef KEY_FIND
 	case KEY_FIND:
 #endif
-	case '/':	search_mode();					break;
+	case '?':
+	case '/':	search_mode(c);					break;
 	case KEY_HOME:
 	case (ctl('a')):
 	case '0':	col_0();					break;
 	case 'A':	u_save(c); last_col(); append_line();		break;
+	case 'B':	linelim = back_word(arg, 1);			break;
 	case 'C':	u_save(c); del_to_end(); append_line();		break;
 	case 'D':	u_save(c); del_to_end();			break;
+	case 'E':	linelim = for_word(arg, 1, 1, 0);		break;
 	case 'F':	linelim = find_char(arg, -1);			break;
 	case 'G':	if (histp > 0) histp = lasthist; for_hist();	break;
 	case 'I':	u_save(c); col_0(); insert_mode();		break;
+	case 'N':	search_again(true);				break;
 	case 'P':	u_save(c);
 			ins_string(putbuf);
 			linelim = back_line(1);				break;
 	case 'R':	u_save(c); replace_mode();			break;
 	case 'T':	linelim = to_char(arg, -1);			break;
+	case 'W':	linelim = for_word(arg, 0, 1, 0);		break;
 	case 'X':	u_save(c); back_space();			break;
 	case 'Y':	yank_chars(linelim, strlen(line), 0);		break;
 	case 'a':	u_save(c); append_line();			break;
-	case 'b':	linelim = back_word(arg);			break;
+	case 'b':	linelim = back_word(arg, 0);			break;
 	case 'c':	u_save(c); yank_cmd(1, 1); insert_mode();	break;
 	case 'd':	u_save(c); yank_cmd(1, 0);			break;
-	case 'e':	linelim = for_word(arg, 1, 0);			break;
+	case 'e':	linelim = for_word(arg, 1, 0, 0);		break;
 	case 'f':	linelim = find_char(arg, 1);			break;
 	case KEY_LEFT:
 	case (ctl('b')):
@@ -177,7 +226,7 @@ write_line(int c)
 	case (ctl('f')):
 	case ' ':
 	case 'l':	linelim = for_line(arg, 0);			break;
-	case 'n':	search_again();					break;
+	case 'n':	search_again(false);				break;
 	case 'p':	u_save(c);
 			linelim = for_line(1, 1);
 			ins_string(putbuf);
@@ -187,7 +236,7 @@ write_line(int c)
 	case 's':	u_save(c); del_in_line(arg, 0); insert_mode();	break;
 	case 't':	linelim = to_char(arg, 1);			break;
 	case 'u':	restore_it();					break;
-	case 'w':	linelim = for_word(arg, 0, 0);			break;
+	case 'w':	linelim = for_word(arg, 0, 0, 0);		break;
 	case KEY_DC:
 	case 'x':	u_save(c); del_in_line(arg, 1);			break;
 	case 'y':	yank_cmd(0, 0);					break;
@@ -206,49 +255,86 @@ write_line(int c)
 	case (ctl('v')):	toggle_navigate_mode();			break;
 	case KEY_LEFT:
 	case (ctl('b')):	if (numeric_field) {
-				    c = craction;
-				    craction = 0;
-				    cr_line();
-				    numeric_field = 0;
-				    craction = c;
-				    backcol(1);
-				} else
-				linelim = back_line(arg);		break;
+				    if (linelim == strlen(line) &&
+					    (line[linelim - 1] == '+' ||
+					     line[linelim - 1] == '-')) {
+					toggle_navigate_mode();
+					backcol(1);
+				    } else {
+					c = craction;
+					craction = 0;
+					cr_line();
+					craction = c;
+					backcol(1);
+				    }
+				} else {
+				    linelim = back_line(arg);
+				    istart = linelim;
+				}   break;
 	case KEY_RIGHT:
 	case (ctl('f')):	if (numeric_field) {
-				    c = craction;
-				    craction = 0;
-				    cr_line();
-				    numeric_field = 0;
-				    craction = c;
-				    forwcol(1);
-				} else
-				linelim = for_line(arg, 1);		break;
+				    if (linelim == strlen(line) &&
+					    (line[linelim - 1] == '+' ||
+					     line[linelim - 1] == '-')) {
+					toggle_navigate_mode();
+					forwcol(1);
+				    } else {
+					c = craction;
+					craction = 0;
+					cr_line();
+					craction = c;
+					forwcol(1);
+				    }
+				} else {
+				    linelim = for_line(arg, 1);
+				    istart = linelim;
+				}   break;
 	case KEY_DOWN:
 	case (ctl('n')):	if (numeric_field) {
-				    c = craction;
-				    craction = 0;
-				    cr_line();
-				    numeric_field = 0;
-				    craction = c;
-				    forwrow(1);
-				} else
-				for_hist();				break;
+				    if (linelim == strlen(line) &&
+					    (line[linelim - 1] == '+' ||
+					     line[linelim - 1] == '-')) {
+					toggle_navigate_mode();
+					forwrow(1);
+				    } else {
+					c = craction;
+					craction = 0;
+					cr_line();
+					craction = c;
+					forwrow(1);
+				    }
+				} else {
+				    for_hist();
+				    istart = linelim;
+				}   break;
 	case KEY_UP:
 	case (ctl('p')):	if (numeric_field) {
-				    c = craction;
-				    craction = 0;
-				    cr_line();
-				    numeric_field = 0;
-				    craction = c;
-				    backrow(1);
-				} else
-				back_hist();				break;
+				    if (linelim == strlen(line) &&
+					    (line[linelim - 1] == '+' ||
+					     line[linelim - 1] == '-')) {
+					toggle_navigate_mode();
+					backrow(1);
+				    } else {
+					c = craction;
+					craction = 0;
+					cr_line();
+					craction = c;
+					backrow(1);
+				    }
+				} else {
+				    back_hist();
+				    istart = linelim;
+				}   break;
 	case KEY_HOME:
 	case (ctl('a')):	col_0();				break;
 	case KEY_END:
 	case (ctl('e')):	last_col();				break;
-	case ESC:		edit_mode();				break;
+	case ESC:		ins_in_line(0);
+				edit_mode();				break;
+	/* '\035' is ^], which expands abbreviations without inserting another
+	 * character in the line
+	 */
+	case '\035':		if (linelim > 0) doabbrev();		break;
 	default:		ins_in_line(c);				break;
 	}
     } else if (mode == SEARCH_MODE) {
@@ -256,7 +342,12 @@ write_line(int c)
 	case KEY_BACKSPACE:
 	case (ctl('h')):	back_space();				break;
 	case (ctl('m')):	search_hist();				break;
-	case ESC:		edit_mode();				break;
+	case ESC:		ins_in_line(0);
+				edit_mode();				break;
+	/* '\035' is ^], which expands abbreviations without inserting another
+	 * character in the line
+	 */
+	case '\035':		if (linelim > 0) doabbrev();		break;
 	default:		ins_in_line(c);				break;
 	}
     } else if (mode == REP_MODE) {
@@ -280,14 +371,43 @@ write_line(int c)
 	case (ctl('i')):	if (!showrange) {
 				    toggle_navigate_mode();
 				    startshow();
+				} else if (linelim == strlen(line) &&
+					(line[linelim - 1] == '+' ||
+					line[linelim - 1] == '-' ||
+					(line[linelim - 1] == ' ' &&
+					 line[linelim - 2] == '='))) {
+				    ins_string("@sum(");
+				    showdr();
+				    ins_in_line(')');
 				} else {
 				    showdr();
 				    ins_in_line(' ');
 				} 					break;
+	case ' ':		if (showrange) {
+				    showdr();
+				    ins_in_line(' ');
+				    toggle_navigate_mode();
+				} else {
+				    forwcol(arg);
+				} 					break;
 	case '+':
 	case '-':		if (!showrange) {
 				    ins_string(v_name(currow, curcol));
 				    ins_in_line(c);
+				} else if (linelim == strlen(line) &&
+					(line[linelim - 1] == '+' ||
+					line[linelim - 1] == '-' ||
+					(line[linelim - 1] == ' ' &&
+					 line[linelim - 2] == '='))) {
+				    ins_string("@sum(");
+				    showdr();
+				    ins_in_line(')');
+				    ins_in_line(c);
+				    toggle_navigate_mode();
+				} else {
+				    showdr();
+				    ins_in_line(')');
+				    ins_in_line(c);
 				}					break;
 	case (ctl('m')):	if (!showrange) {
 				    ins_string(v_name(currow, curcol));
@@ -296,8 +416,7 @@ write_line(int c)
 				    toggle_navigate_mode();
 				    cr_line();
 				}					break;
-	case 'v':
-	case (ctl('v')):	{   /* insert variable value */
+	case 'v':	{   /* insert variable value */
 				    struct ent *p = *ATBL(tbl, currow, curcol);
 				    char temp[100];
 
@@ -338,7 +457,6 @@ write_line(int c)
 	case KEY_LEFT:
 	case 'h':		backcol(arg);				break;
 	case KEY_RIGHT:
-	case ' ':
 	case 'l':		forwcol(arg);				break;
 	case KEY_DOWN:
 	case (ctl('n')):
@@ -348,6 +466,7 @@ write_line(int c)
 	case 'k':		backrow(arg);				break;
 	case 'q':
 	case ctl('g'):
+	case (ctl('v')):
 	case ESC:		toggle_navigate_mode();
 				showrange = 0;				break;
 	case 'H':		backcol(curcol - stcol + 2);
@@ -379,62 +498,29 @@ write_line(int c)
 				FullUpdate++;
 				}
 									break;
-	case 'L':
-				forwcol(lcols -(curcol-stcol)+1);
-									break;
+	case 'L':		forwcol(lcols - (curcol - stcol) + 1);	break;
 	case (ctl('a')):
-	case KEY_HOME:
-				remember(0);
-				currow = 0;
-				curcol = 0;
-				rowsinrange = 1;
-				colsinrange = fwidth[curcol];
-				remember(1);
-				FullUpdate++;				break;
-	case '0':
-				remember(0);
-				curcol = 0;
-		   		rowsinrange = 1;
-		   		colsinrange = fwidth[curcol];
-				remember(1);				break;
-	case '$':		{
-				register struct ent *p;
+	case KEY_HOME:		gohome();				break;
+	case '0':		leftlimit();				break;
+	case '$':		rightlimit();				break;
+	case '^':		gototop();				break;
+	case '#':		gotobottom();				break;
+	case 'o':		if (showrange) {
+				    int r = currow;
+				    int c = curcol;
 
-				remember(0);
-				curcol = maxcols - 1;
-				while (!VALID_CELL(p, currow, curcol) &&
-					curcol > 0)
-				    curcol--;
-				rowsinrange = 1;
-				colsinrange = fwidth[curcol];
-				remember(1);				break;
-				}
-	case '^':
-				remember(0);
-				currow = 0;
-				rowsinrange = 1;
-				colsinrange = fwidth[curcol];
-				remember(1);				break;
-	case '#':		{
-				register struct ent *p;
-
-				remember(0);
-				currow = maxrows - 1;
-				while (!VALID_CELL(p, currow, curcol) &&
-					currow > 0)
-				    currow--;
-				rowsinrange = 1;
-				colsinrange = fwidth[curcol];
-				remember(1);				break;
-				}
-	case 'm':
-				markcell();				break;
-	case '`': case '\'':
-				dotick(c);				break;
-	case '*':
-				gotonote();				break;
-	case 'g':
-				dogoto();				break;
+				    currow = showsr;
+				    showsr = r;
+				    curcol = showsc;
+				    showsc = c;
+				    rowsinrange = 1;
+				    colsinrange = fwidth[curcol];
+				} 					break;
+	case 'm':		markcell();				break;
+	case '`': case '\'':	dotick(c);				break;
+	case '*':		if (nmgetch() == '*') gotonote();	break;
+	case 'g':		dogoto();				break;
+	case 'n':		go_last();				break;
 	case 'w':		{
 				register struct ent *p;
 
@@ -484,10 +570,9 @@ write_line(int c)
 				colsinrange = fwidth[curcol];
 				break;
 				}
-	case 'C':		if (braille) {
+	case 'C':		if (braille)
 				    braillealt ^= 1;
-				    break;
-				}
+				break;
 	}
     }
 }
@@ -495,8 +580,8 @@ write_line(int c)
 void
 edit_mode()
 {
-    mode = EDIT_MODE;
     mode_ind = 'e';
+    mode = EDIT_MODE;
     if (linelim < 0)	/* -1 says stop editing, ...so we still aren't */
 	return;
     numeric_field = 0;
@@ -508,16 +593,29 @@ insert_mode()
 {
     mode_ind = 'i';
     mode = INSERT_MODE;
+    istart = linelim;
 }
 
 static void
-search_mode()
+search_mode(char sind)
 {
-    line[0] = '/';
-    line[1] = '\0';
-    linelim = 1;
-    mode_ind = '/';
-    mode = SEARCH_MODE;
+    if (search_ind == ' ') {
+	/*
+	 * This back and forth movement through the history is just a quick
+	 * way to force the current command to be saved in history[0].histline,
+	 * allocating space for it if necessary.  The command will be copied
+	 * back into line by search_hist() before the search is done. - CRM
+	 */
+	back_hist();
+	for_hist();
+	line[0] = '\0';
+	linelim = 0;
+	mode_ind = 'i';
+	search_ind = sind;
+	search_dir = sind == '/' ? 1 : 0;
+	mode = SEARCH_MODE;
+	istart = linelim;
+    }
 }
 
 static void
@@ -613,9 +711,6 @@ startshow()
 
 /* insert the range we defined by moving around the screen, see startshow() */
 
-#define SHOWROWS 2
-#define SHOWCOLS 3
-
 void
 showdr()
 {
@@ -775,11 +870,17 @@ restore_it()
 static void
 stop_edit()
 {
-    showrange = 0;
-    numeric_field = 0;
-    linelim = -1;
-    (void) move(1, 0);
-    (void) clrtoeol();
+    if (search_ind != ' ') {
+	search_ind = ' ';
+	(void) strcpy(line, history[0].histline);
+	write_line('G');
+    } else {
+	showrange = 0;
+	numeric_field = 0;
+	linelim = -1;
+	(void) move(1, 0);
+	(void) clrtoeol();
+    }
 }
 
 /*
@@ -811,7 +912,7 @@ for_line(int arg, int stop_null)
  */
 
 static int
-for_word(int arg, int end_word, int stop_null)
+for_word(int arg, int end_word, int big_word, int stop_null)
 {
     register int c;
     register int cpos;
@@ -831,7 +932,10 @@ for_word(int arg, int end_word, int stop_null)
 		continue;
 	}
 
-	if (istext(line[cpos])) {
+	if (big_word)
+	    while ((c = line[cpos]) && c != ' ')
+		cpos++;
+	else if (istext(line[cpos])) {
 	    while ((c = line[cpos]) && istext(c)) 
 		cpos++;
 	} else {
@@ -861,7 +965,7 @@ back_line(int arg)
 }
 
 static int
-back_word(int arg)
+back_word(int arg, int big_word)
 {
     register int c;
     register int cpos;
@@ -883,12 +987,15 @@ back_word(int arg)
 	}
 
 	/* Skip across the word - goes 1 too far */
-	if (istext(line[cpos])) {
+	if (big_word)
+	    while (cpos > 0 && (c = line[cpos]) && c != ' ')
+		--cpos;
+	else if (istext(line[cpos])) {
 	    while (cpos > 0 && (c = line[cpos]) && istext(c)) 
-		    --cpos;
+		--cpos;
 	} else {
 	    while (cpos > 0 && (c = line[cpos]) && !istext(c) && c != ' ')
-		    --cpos;
+		--cpos;
 	}
 
 	/* We are done - fix up the one too far */
@@ -927,18 +1034,26 @@ del_in_line(int arg, int back_null)
 void
 ins_in_line(int c)
 {
-    register int i, len;
+    int i, len;
+    static int inabbr;
 
     if (c < 256) {
-	if (linelim < 0) {
+	if (linelim < 0 && c > 0) {
 	    *line = '\0';
 	    linelim = 0;
 	}
-	len = strlen(line);
-	for (i = len; i >= linelim; --i)
-	    line[i+1] = line[i];
-	line[linelim++] = c;
-	line[len+1] = '\0';
+	if (!inabbr && linelim > 0 && !(isalpha(c) || isdigit(c) || c == '_')) {
+	    inabbr++;
+	    doabbrev();
+	    inabbr--;
+	}
+	if (c > 0) {
+	    len = strlen(line);
+	    for (i = len; i >= linelim; --i)
+		line[i+1] = line[i];
+	    line[linelim++] = c;
+	    line[len+1] = '\0';
+	}
     }
 }
 
@@ -949,6 +1064,52 @@ ins_string(char *s)
 	ins_in_line(*s++);
 }
 
+void
+doabbrev()
+{
+    int len, pos;
+    struct abbrev *a;
+    struct abbrev *prev;
+    
+    if (!(isalpha(line[linelim-1]) || isdigit(line[linelim-1]) ||
+	    line[linelim-1] == '_') || !(mode == INSERT_MODE ||
+	    mode == SEARCH_MODE) || istart >= linelim)
+	return;
+
+    pos = linelim - 2;
+    if (isalpha(line[pos]) || isdigit(line[pos]) || line[pos] == '_') {
+	for (; pos >= istart; pos--)
+	    if (!(isalpha(line[pos]) || isdigit(line[pos]) || line[pos] == '_'))
+		break;
+    } else if (line[pos] != ' ')
+	for (; pos >= istart; pos--)
+	    if (isalpha(line[pos]) || isdigit(line[pos]) || line[pos] == '_' ||
+		    line[pos] == ' ')
+		break;
+    pos++;
+
+    if (istart && pos == istart) {
+	if (isalpha(line[pos]) || isdigit(line[pos]) || line[pos] == '_') {
+	    if (isalpha(line[--pos]) || isdigit(line[pos]) || line[pos] == '_')
+		return;
+	} else {
+	    if (!(isalpha(line[--pos]) || isdigit(line[pos]) ||
+		    line[pos] == '_' || line[pos] == ' '))
+		return;
+	}
+	pos++;
+    }
+
+    len = linelim - pos;
+    if (len && (a = find_abbr(line + pos, len, &prev))) {
+	if (len > 1 || pos == 0 || line[pos-1] == ' ') {
+	    linelim = pos;
+	    del_in_line(len, 0);
+	    ins_string(a->exp);
+	}
+    }
+}
+
 static void
 append_line()
 {
@@ -1025,6 +1186,8 @@ back_space()
 	linelim = back_line(1);
 	del_in_line(1, 1);
     }
+    if (linelim < istart)
+	istart = linelim;
 }
 
 /* Setting change to 1 makes `w' act like `e' so that `cw' will act like
@@ -1038,7 +1201,7 @@ get_motion(int change)
     int arg2 = 0;
 
     c = vigetch();
-    if (c == 0) {
+    if (c == '0') {
 	savedot(c);
 	return (0);
     }
@@ -1055,17 +1218,20 @@ get_motion(int change)
     }
     switch (c) {
 	case '$':	return (strlen(line));
-	case 'b':	return (back_word(arg));
+	case 'b':	return (back_word(arg, 0));
+	case 'B':	return (back_word(arg, 1));
 	case 'c':	return (change ? -1 : linelim);
 	case 'd':	return (!change ? -1 : linelim);
-	case 'e':	return (for_word(arg, 1, 1) + 1);
+	case 'e':	return (for_word(arg, 1, 0, 1) + 1);
+	case 'E':	return (for_word(arg, 1, 1, 1) + 1);
 	case 'f':	return ((c = find_char(arg, 1)) == linelim ? c : c + 1);
 	case 'F':	return (find_char(arg, -1));
 	case 'h':	return (back_line(arg));
 	case 'l':	return (for_line(arg, 1));
 	case 't':	return ((c = to_char(arg, 1)) == linelim ? c : c + 1);
 	case 'T':	return (to_char(arg, -1));
-	case 'w':	return (for_word(arg, change, 1) + change);
+	case 'w':	return (for_word(arg, change, 0, 1) + change);
+	case 'W':	return (for_word(arg, change, 1, 1) + change);
 	default:	return (linelim);
     }
 }
@@ -1116,56 +1282,146 @@ cr_line()
 {
     struct frange *fr;
 
+    ins_in_line(0);
     insert_mode();
-    if (linelim != -1) {
-	save_hist();
-	nosavedot = 1;
+    numeric_field = 0;
+    if (linelim == -1) {	/* '\n' alone will put you into insert mode */
+    	*line = '\0';		/* unless numeric and craction are both	set */
 	linelim = 0;
-	(void) yyparse();
-	showrange = 0;
-	linelim = -1;
-	if (cellassign) {
-	    cellassign = 0;
-	    switch (craction) {
-		case CRROWS:
-		    if ((rowlimit >= 0) && (currow >= rowlimit)) {
-			forwcol(1);
-			currow = 0;
-		    } else {
-			if (autoinsert && (fr = find_frange(currow, curcol))) {
-			    forwrow(1);
-			    if (currow > fr->ir_right->row) {
-				backrow(1);
-				insertrow(1, 1);
-			    }
-			} else
-			    forwrow(1);
-		    }
-		    break;
-		case CRCOLS:
-		    if ((collimit >= 0) && (curcol >= collimit)) {
+	if (numeric && craction)
+	    cellassign = 1;
+	else
+	    return;
+    }
+    save_hist();
+    nosavedot = 1;
+    linelim = 0;
+    (void) yyparse();
+    showrange = 0;
+    linelim = -1;
+    if (cellassign) {
+	cellassign = 0;
+	switch (craction) {
+	    case CRROWS:
+		if ((rowlimit >= 0) && (currow >= rowlimit)) {
+		    forwcol(1);
+		    currow = 0;
+		} else {
+		    if ((fr = find_frange(currow, curcol))) {
 			forwrow(1);
-			curcol = 0;
-		    } else {
-			if (autoinsert && (fr = find_frange(currow, curcol))) {
-			    forwcol(1);
-			    if (curcol > fr->ir_right->col) {
-				backcol(1);
+			if (currow > fr->ir_right->row) {
+			    backrow(1);
+			    if (autowrap) {
+				forwcol(1);
+				currow = fr->ir_left->row;
+				if (row_hidden[currow])
+				    forwrow(1);
+				if (curcol > fr->ir_right->col) {
+				    backcol(1);
+				    if (autoinsert)
+					insertcol(1, 1);
+				    else {
+					currow = fr->ir_right->row;
+					if (row_hidden[currow])
+					    backrow(1);
+				    }
+				}
+			    } else if (autoinsert)
+				insertrow(1, 1);
+			}
+		    } else
+			forwrow(1);
+		}
+		break;
+	    case CRCOLS:
+		if ((collimit >= 0) && (curcol >= collimit)) {
+		    forwrow(1);
+		    curcol = 0;
+		} else {
+		    if ((fr = find_frange(currow, curcol))) {
+			forwcol(1);
+			if (curcol > fr->ir_right->col) {
+			    backcol(1);
+			    if (autowrap) {
+				forwrow(1);
+				curcol = fr->ir_left->col;
+				if (col_hidden[curcol])
+				    forwcol(1);
+				if (currow > fr->ir_right->row) {
+				    backrow(1);
+				    if (autoinsert)
+					insertrow(1, 1);
+				    else {
+					curcol = fr->ir_right->col;
+					if (col_hidden[curcol])
+					    backcol(1);
+				    }
+				}
+			    } else if (autoinsert)
 				insertcol(1, 1);
-			    }
-			} else
-			    forwcol(1);
-		    }
-		    break;
-		default:
-		    break;
-	    }
+			}
+		    } else
+			forwcol(1);
+		}
+		break;
+	    default:
+		break;
 	}
     }
-    else {	/* '\n' alone will put you into insert mode */
-    	*line = '\0';
-	linelim = 0;
+}
+
+void
+doshell()
+{
+    /*
+    *  "! command"  executes command
+    *  "!"	forks a shell
+    *  "!!" repeats last command
+    */
+#if VMS || MSDOS
+    error("Not implemented on VMS or MS-DOS");
+#else /* VMS */
+    char *shl;
+    int pid, temp;
+    char cmd[MAXCMD];
+    static char lastcmd[MAXCMD];
+
+    if (!(shl = getenv("SHELL")))
+	shl = "/bin/sh";
+
+    deraw(1);
+    (void) fputs("! ", stdout);
+    (void) fflush(stdout);
+    (void) fgets(cmd, MAXCMD, stdin);
+    cmd[strlen(cmd) - 1] = '\0';	/* clobber \n */
+    if (strcmp(cmd,"!") == 0)		/* repeat? */
+	(void) strcpy(cmd, lastcmd);
+    else
+	(void) strcpy(lastcmd, cmd);
+
+    if (modflg) {
+	(void) puts("[No write since last change]");
+	(void) fflush(stdout);
     }
+
+    if (!(pid = fork())) {
+	(void) signal(SIGINT, SIG_DFL);  /* reset */
+	if (strlen(cmd))
+	    (void) execl(shl, shl, "-c", cmd, NULL);
+	else
+	    (void) execl(shl, shl, NULL);
+	exit (-127);
+    }
+
+    while (pid != wait(&temp));
+
+    (void) printf("Press any key to continue ");
+    fflush(stdout);
+    cbreak();
+    (void)getch();
+    goraw();
+    clear();
+#endif /* VMS */
 }
 
 /* History functions */
@@ -1182,21 +1438,45 @@ save_hist()
 	if (lasthist > endhist)
 	    endhist = lasthist;
 
-	if (history[lasthist].len < strlen(line)+1) {
-	    history[lasthist].len = strlen(line)+40;
+	if (history[lasthist].len < strlen(line) + 1) {
+	    history[lasthist].len = strlen(line) + 40;
 	    history[lasthist].histline = scxrealloc(history[lasthist].histline,
 		    history[lasthist].len);
 	}
 	(void) strcpy(history[lasthist].histline, line);
+	histsessionnew++;
     }
     if (history[0].histline) {
 	scxfree(history[0].histline);
-	history[0].histline = (void *)0;
+	history[0].histline = NULL;
 	history[0].len = 0;
     }
     histp = 0;
 }
 
+static void
+for_hist()
+{
+    if (histp == 0) {
+	last_col();
+    	return;
+    }
+
+    if (histp == lasthist)
+	histp = 0;
+    else
+	histp = histp % endhist + 1;
+
+    if (lasthist >= 0) {
+	(void) strcpy(line, history[histp].histline);
+	last_col();
+    }
+    if (histp)
+	error("History line %d", endhist - lasthist + histp);
+    else
+	error("");
+}
+
 static void
 back_hist()
 {
@@ -1220,49 +1500,137 @@ back_hist()
     	(void) strcpy(line, history[histp].histline);
 	last_col();
     }
+    if (histp)
+	error("History line %d", endhist - lasthist + histp);
+    else
+	error("");
 }
 
 static void
 search_hist()
 {
+#ifdef RECOMP
+    char	*tmp;
+#endif
+#if !defined(REGCOMP) && !defined(RE_COMP) && !defined(REGCMP)
     static	unsigned lastsrchlen = 0;
+#endif
 
-    if(linelim < 1) {
+    if (linelim < 1) {
 	linelim = 0;
 	edit_mode();
 	return;
     }
 
-    if (strlen(line)+1 > lastsrchlen) {
-    	lastsrchlen = strlen(line)+40;
+#ifdef REGCOMP
+    if (last_search)
+	regfree(last_search);
+    else
+	last_search = &preg;
+    if ((errcode = regcomp(last_search, line, REG_EXTENDED))) {
+	char *tmp = scxmalloc((size_t)160);
+	regerror(errcode, last_search, tmp, sizeof(tmp));
+	error(tmp);
+	scxfree(tmp);
+	return;
+    }
+#else
+#ifdef RE_COMP
+    if ((tmp = re_comp(line)) != NULL) {
+	error(tmp);
+	return;
+    }
+#else
+#ifdef REGCMP
+    free(last_search);
+    if ((last_search = regcmp(line, NULL)) == NULL) {
+	error("Invalid search string");
+	return;
+    }
+#else
+    if (strlen(line) + 1 > lastsrchlen) {
+    	lastsrchlen = strlen(line) + 40;
 	last_search = scxrealloc(last_search, lastsrchlen);
     }
-    (void)strcpy(last_search, line+1);
-    search_again();
-    mode = EDIT_MODE;
+    (void) strcpy(last_search, line);
+#endif
+#endif
+#endif
+    (void) strcpy(line, history[0].histline);
+    search_again(false);
+    if (mode != EDIT_MODE) edit_mode();
+    search_ind = ' ';
 }
 
 static void
-search_again()
+search_again(bool reverse)
 {
+    int prev_match;
     int found_it;
-    int do_next;
-    int prev_histp;
+#if !defined(REGCOMP) && !defined(RE_COMP) && !defined(REGCMP)
     char *look_here;
+    int do_next;
+#endif
 
-    prev_histp = histp;
+#ifdef REGCOMP
+    if ((last_search == NULL))
+	return;
+#else
+#ifndef RE_COMP
     if ((last_search == NULL) || (*last_search == '\0'))
 	return;
+#endif
+#endif
+    prev_match = histp > 0 ? histp : 0;
+    error("");
 
     do {
-	back_hist();
-	if (prev_histp == histp)
+	if (lasthist > 0) {
+	    if (!(search_dir ^ reverse) && histp != lasthist)
+		if (histp <= 0) {
+		    histp = ((lasthist + 1) % endhist);
+		    (void) strcpy(line, history[histp].histline);
+		} else
+		    for_hist();
+	    else if ((search_dir ^ reverse) && histp != ((lasthist + 1) % endhist))
+		back_hist();
+	    else {
+		histp = 0;
+		(void) strcpy(line, history[0].histline);
+		last_col();
+	    }
+	} else
 	    break;
-	prev_histp = histp;
+	if (histp == prev_match) {
+	    if (histp <= 0) {
+		error("No matches found");
+		break;
+	    }
+	}
+	if (histp <= 0) {
+	    if (search_dir ^ reverse)
+		back_hist();
+	    else
+		histp = ((lasthist + 1) % endhist);
+		(void) strcpy(line, history[histp].histline);
+	}
+	found_it = 0;
+#ifdef REGCOMP
+	if (regexec(last_search, line, 0, NULL, 0) == 0)
+	    found_it++;
+#else
+#ifdef RE_COMP
+	if (re_exec(line) != 0)
+	    found_it++;
+#else
+#ifdef REGCMP
+	if (regex(last_search, line) != NULL)
+	    found_it++;
+#else
 	look_here = line;
-	found_it = do_next = 0;
+	do_next = 0;
 	while ((look_here = (char *)strchr(look_here, *last_search)) != NULL &&
-						!found_it && !do_next) {
+		!found_it && !do_next) {
 
 	    if (strncmp(look_here, last_search, strlen(last_search)) == 0)
 		found_it++;
@@ -1271,26 +1639,82 @@ search_again()
 	    else
 		do_next++;
 	}
+#endif
+#endif
+#endif
+	if (histp == prev_match)
+	    break;
     } while (!found_it);
+    if (found_it)
+	error("History line %d", endhist - lasthist + histp);
+    else
+	error("No matches found");
+    edit_mode();
+    linelim = strlen(line) - 1;
+}
+
+#if !defined(MSDOS) && defined HISTORY_FILE
+void
+write_hist()
+{
+    int i;
+    FILE *fp, *tmpfp = NULL;
+    char histfile[PATHLEN];
+
+    if (histsessionnew < HISTLEN) {
+	/* write the new history for this session to a tmp file */
+	tmpfp = tmpfile();
+	for (i = 1; i <= histsessionnew; i++) {
+	    histsessionstart = histsessionstart % endhist + 1;
+	    if (history[histsessionstart].len > 40)
+		fprintf(tmpfp, "%s\n", history[histsessionstart].histline);
+	}
+	fseek(tmpfp, 0, SEEK_SET);
+
+	/* re-read the main history, then read back in the saved session hist*/
+	histp = 0;
+	lasthist = 0;
+	endhist = -1;
+	read_hist();
+	readhistfile(tmpfp);
+    }
+
+    /* now write to whole lot out to the proper save file */
+    strcpy(histfile, HISTORY_FILE);
+    if (findhome(histfile) && (fp = fopen(histfile, "w")) != NULL) {
+	for (i = 1; i <= endhist; i++) {
+	    lasthist = lasthist % endhist + 1;
+	    if (history[lasthist].len > 40)
+		fprintf(fp, "%s\n", history[lasthist].histline);
+	}
+	fclose(fp);
+    }
 }
 
 static void
-for_hist()
+readhistfile(FILE *fp)
 {
-    if (histp == 0)
-    	return;
-
-    if (histp == lasthist)
-	histp = 0;
-    else
-	histp = histp % endhist + 1;
-
-    if (lasthist >= 0) {
-	(void) strcpy(line, history[histp].histline);
-	last_col();
+    while (fgets(line, FBUFLEN, fp)) {
+	line[strlen(line)-1] = '\0'; /* chop the \n */
+	save_hist();
     }
+    fclose(fp);
 }
 
+void
+read_hist()
+{
+    FILE *fp;
+    char histfile[PATHLEN];
+
+    strcpy(histfile, HISTORY_FILE);
+    if (findhome(histfile) && (fp = fopen(histfile, "r")) != NULL)
+	readhistfile(fp);
+    histsessionstart = lasthist;
+    histsessionnew = 0;
+}
+#endif
+
 static void
 col_0()
 {
@@ -1402,6 +1826,159 @@ remember(int save)
     }
 }
 
+void
+gohome()
+{
+    struct frange *fr;
+
+    remember(0);
+    if ((fr = find_frange(currow, curcol))) {
+	if (currow >= fr->ir_left->row &&
+		currow <= fr->ir_right->row &&
+		curcol >= fr->ir_left->col &&
+		curcol <= fr->ir_right->col &&
+		(currow > fr->ir_left->row ||
+		curcol > fr->ir_left->col)) {
+	    currow = fr->ir_left->row;
+	    curcol = fr->ir_left->col;
+	} else if (currow > fr->or_left->row ||
+		curcol > fr->or_left->col) {
+	    currow = fr->or_left->row;
+	    curcol = fr->or_left->col;
+	} else {
+	    currow = 0;
+	    curcol = 0;
+	}
+    } else {
+	currow = 0;
+	curcol = 0;
+    }
+    rowsinrange = 1;
+    colsinrange = fwidth[curcol];
+    remember(1);
+    FullUpdate++;
+}
+
+void
+leftlimit()
+{
+    struct frange *fr;
+
+    remember(0);
+    if ((fr = find_frange(currow, curcol))) {
+	if (currow >= fr->ir_left->row &&
+		currow <= fr->ir_right->row &&
+		curcol > fr->ir_left->col &&
+		curcol <= fr->ir_right->col)
+	    curcol = fr->ir_left->col;
+	else if (curcol > fr->or_left->col &&
+		curcol <= fr->or_right->col)
+	    curcol = fr->or_left->col;
+	else
+	    curcol = 0;
+    } else
+	curcol = 0;
+    rowsinrange = 1;
+    colsinrange = fwidth[curcol];
+    remember(1);
+}
+
+void
+rightlimit()
+{
+    register struct ent *p;
+    struct frange *fr;
+
+    remember(0);
+    if ((fr = find_frange(currow, curcol))) {
+	if (currow >= fr->ir_left->row &&
+		currow <= fr->ir_right->row &&
+		curcol >= fr->ir_left->col &&
+		curcol < fr->ir_right->col)
+	    curcol = fr->ir_right->col;
+	else if (curcol >= fr->or_left->col &&
+		curcol < fr->or_right->col)
+	    curcol = fr->or_right->col;
+	else {
+	    curcol = maxcols - 1;
+	    while (!VALID_CELL(p, currow, curcol) &&
+		    curcol > fr->or_right->col)
+		curcol--;
+	    if ((fr = find_frange(currow, curcol)))
+		curcol = fr->or_right->col;
+	}
+    } else {
+	curcol = maxcols - 1;
+	while (!VALID_CELL(p, currow, curcol) && curcol > 0)
+	    curcol--;
+	if ((fr = find_frange(currow, curcol)))
+	    curcol = fr->or_right->col;
+    }
+    rowsinrange = 1;
+    colsinrange = fwidth[curcol];
+    remember(1);
+}
+
+void
+gototop()
+{
+    struct frange *fr;
+
+    remember(0);
+    if ((fr = find_frange(currow, curcol))) {
+	if (curcol >= fr->ir_left->col &&
+		curcol <= fr->ir_right->col &&
+		currow > fr->ir_left->row &&
+		currow <= fr->ir_right->row)
+	    currow = fr->ir_left->row;
+	else if (currow > fr->or_left->row &&
+		currow <= fr->or_right->row)
+	    currow = fr->or_left->row;
+	else
+	    currow = 0;
+    } else
+	currow = 0;
+    rowsinrange = 1;
+    colsinrange = fwidth[curcol];
+    remember(1);
+}
+
+void
+gotobottom()
+{
+    register struct ent *p;
+    struct frange *fr;
+
+    remember(0);
+    if ((fr = find_frange(currow, curcol))) {
+	if (curcol >= fr->ir_left->col &&
+		curcol <= fr->ir_right->col &&
+		currow >= fr->ir_left->row &&
+		currow < fr->ir_right->row)
+	    currow = fr->ir_right->row;
+	else if (currow >= fr->or_left->row &&
+		currow < fr->or_right->row)
+	    currow = fr->or_right->row;
+	else {
+	    currow = maxrows - 1;
+	    while (!VALID_CELL(p, currow, curcol) &&
+		    currow > fr->or_right->row)
+		currow--;
+	    if ((fr = find_frange(currow, curcol)))
+		currow = fr->or_right->row;
+	}
+    } else {
+	currow = maxrows - 1;
+	while (!VALID_CELL(p, currow, curcol) && currow > 0)
+	    currow--;
+	if ((fr = find_frange(currow, curcol)))
+	    currow = fr->or_right->row;
+    }
+    rowsinrange = 1;
+    colsinrange = fwidth[curcol];
+    remember(1);
+}
+
 static void
 dogoto()
 {
